// Copyright 2022 Haute école d'ingénierie et d'architecture de Fribourg
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/****************************************************************************
 * @file gear_device.cpp
 * @author Serge Ayer <serge.ayer@hefr.ch>
 * @author Rémi Heredero <remi@heredero.ch>
 * @author Yann Sierro <yannsierro.pro@gmail.com>
 *
 * @brief Gear Device implementation (static scheduling)
 *
 * @date 2023-11-17
 * @version 1.1.0
 ***************************************************************************/

#include "gear_device.hpp"

#include "bike_system.hpp"

// from disco_h747i/wrappers

#include "joystick.hpp"
#include "mbed_atomic.h"
#include "mbed_trace.h"

#if MBED_CONF_MBED_TRACE_ENABLE
#define TRACE_GROUP "GearDevice"
#endif  // MBED_CONF_MBED_TRACE_ENABLE

namespace multi_tasking {

GearDevice::GearDevice(Mail<gearMail_t, 16>* mailBox, Timer& timer)
    : _mailBox(mailBox), _timer(timer) {
    disco::Joystick::getInstance().setUpCallback(callback(this, &GearDevice::onUp));
    disco::Joystick::getInstance().setDownCallback(callback(this, &GearDevice::onDown));
}

void GearDevice::onUp() {
    if (_currentGear < bike_computer::kMaxGear) {
        core_util_atomic_incr_u8(&_currentGear, 1);
        sendMail(core_util_atomic_load_u8(&_currentGear));
    }
}

void GearDevice::onDown() {
    if (_currentGear > bike_computer::kMinGear) {
        core_util_atomic_decr_u8(&_currentGear, 1);
        sendMail(core_util_atomic_load_u8(&_currentGear));
    }
}

void GearDevice::sendMail(uint32_t data) {
    if (_mailBox != nullptr) {
        gearMail_t* currentGear = _mailBox->try_alloc();
        if (currentGear != nullptr) {
            currentGear->gear     = data;
            currentGear->callTime = _timer.elapsed_time();
            _mailBox->put(currentGear);
        }
    }
}

}  // namespace multi_tasking
