// Copyright 2022 Haute école d'ingénierie et d'architecture de Fribourg
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/****************************************************************************
 * @file pedal_device.hpp
 * @author Serge Ayer <serge.ayer@hefr.ch>
 * @author Rémi Heredero <remi@heredero.ch>
 * @author Yann Sierro <yannsierro.pro@gmail.com>
 *
 * @brief Pedal System header file (static scheduling)
 *
 * @date 2023-11-17
 * @version 1.1.0
 ***************************************************************************/

#pragma once

#include "constants.hpp"
#include "mbed.h"

namespace multi_tasking {

typedef struct pedalMail {
    uint8_t step;
    std::chrono::microseconds callTime;
} pedalMail_t;

class PedalDevice {
   public:
    explicit PedalDevice(Mail<pedalMail_t, 16>* mailBox,  // NOLINT (runtime/references)
                         Timer& timer);                   // NOLINT (runtime/references)

    // make the class non copyable
    PedalDevice(PedalDevice&)            = delete;
    PedalDevice& operator=(PedalDevice&) = delete;

    // method called for updating the bike system
    static std::chrono::milliseconds getCurrentRotationTime(uint32_t step);

   private:
    // private methods
    void onLeft();
    void onRight();
    void increaseRotationSpeed();
    void decreaseRotationSpeed();

    void sendMail(uint32_t data);
    // data members
    volatile uint32_t _currentStep = static_cast<uint32_t>(
        (bike_computer::kInitialPedalRotationTime - bike_computer::kMinPedalRotationTime)
            .count() /
        bike_computer::kDeltaPedalRotationTime.count());

    Mail<pedalMail_t, 16>* _mailBox;
    Timer& _timer;
};

}  // namespace multi_tasking
